/*
Copyright (c) <2010> <Matthew Hausknecht, Peter Stone>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
*/

//-*-c++-*-
#ifndef INCLUDED_InvertedKick_h_
#define INCLUDED_InvertedKick_h_

#include "Behaviors/StateNode.h"
#include "Motion/MotionManager.h"
#include "Motion/MotionSequenceMC.h"
#include "Shared/mathutils.h"

using namespace mathutils;

class InvertedKick : public StateNode {

public:
  InvertedKick()
    : StateNode("InvertedKick","InvertedKick")
  {}

  InvertedKick(const std::string& nm)
    : StateNode("InvertedKick",nm)
  {}

protected:
  InvertedKick(const std::string &class_name, const std::string &node_name)
    : StateNode(class_name,node_name)
  {}
	
	
public:
  virtual void DoStart() {
    StateNode::DoStart(); // do this first (required)

    erouter->addListener(this,EventBase::textmsgEGID);
  }

  /* Attempts to modulate the power of a kick in order to make the ball
   * go the desired distance in centimeters.
   */
  void variableKick (SharedObject<MediumMotionSequenceMC> mseq_mc, int desiredDistCM) {
    //Quadratic: -.33x^2 + 17.7174x + 160.0987
    //Max @ x=26.84 with value 397.91
    if (desiredDistCM < 0)
      return;
    
    int numFrames;
    if (desiredDistCM >= 398) { // Give hardest known kick
      numFrames = (int)round(26.84);
    } else if (desiredDistCM >= 160) { // Use Left Side of Quadratic
      numFrames = (int) round(-(sqrt(15.0)*sqrt(87437.0-220.0*desiredDistCM)-885.0)/33.0);
    } else { // Use Right Side of Quadratic
      numFrames = (int) round((sqrt(15.0)*sqrt(87437.0-220.0*desiredDistCM)+885.0)/33.0);
    }

    std::cout << "Using " << numFrames << " Frames for kick" << std::endl;
    
    setPose(mseq_mc,24,-40,0,-20,-50,0,0,60,-8,30,60,-8,30,-45,6,75,-45,6,75);
    setPose(mseq_mc,24,0,-75,-20,0,0,0,-10,-8,30,-10,-8,30,-60,-5,120,-60,-5,120);
    setPose(mseq_mc,16,0,-75,-20,0,0,0,-10,-8,30,-10,-8,30,-60,0,120,-60,0,120);
    setPose(mseq_mc,numFrames,-35,35,-20,0,0,0,75,-7,15,75,-7,15,30,0,120,30,0,120);
    setPose(mseq_mc,24,-35,35,-20,0,0,0,75,-7,15,75,-7,15,30,0,120,30,0,120);
    setPose(mseq_mc,48,0,0,-20,0,0,0,-5,10,115,-5,10,115,-30,6,104,-30,6,104);
  }

  void setPose (SharedObject<MediumMotionSequenceMC> mseq_mc, int numFrames,
		int neckTilt, int headPan, int headTilt,
		int mouth, int tailTilt, int tailPan,
		int frontLeftRotator, int frontLeftAbductor, int frontLeftKnee,
		int frontRightRotator, int frontRightAbductor, int frontRightKnee,
		int backLeftRotator, int backLeftAbductor, int backLeftKnee,
		int backRightRotator, int backRightAbductor, int backRightKnee) {
    // Number of frames to get into this pose
    mseq_mc->advanceTime(8*numFrames);
    
    // Head
    mseq_mc->setOutputCmd(HeadOffset+TiltOffset, RAD(neckTilt));
    mseq_mc->setOutputCmd(HeadOffset+PanOffset, RAD(headPan));
    mseq_mc->setOutputCmd(HeadOffset+NodOffset, RAD(headTilt));

    // Mouth & Tail
    mseq_mc->setOutputCmd(MouthOffset, RAD(mouth));
    mseq_mc->setOutputCmd(TailOffset+TiltOffset, RAD(tailTilt));
    mseq_mc->setOutputCmd(TailOffset+PanOffset, RAD(tailPan));

    // Legs
    mseq_mc->setOutputCmd(LFrLegOffset+RotatorOffset, RAD(frontLeftRotator));
    mseq_mc->setOutputCmd(LFrLegOffset+ElevatorOffset, RAD(frontLeftAbductor));
    mseq_mc->setOutputCmd(LFrLegOffset+KneeOffset, RAD(frontLeftKnee));
    
    mseq_mc->setOutputCmd(RFrLegOffset+RotatorOffset, RAD(frontRightRotator));
    mseq_mc->setOutputCmd(RFrLegOffset+ElevatorOffset, RAD(frontRightAbductor));
    mseq_mc->setOutputCmd(RFrLegOffset+KneeOffset, RAD(frontRightKnee));

    mseq_mc->setOutputCmd(LBkLegOffset+RotatorOffset, RAD(backLeftRotator));
    mseq_mc->setOutputCmd(LBkLegOffset+ElevatorOffset, RAD(backLeftAbductor));
    mseq_mc->setOutputCmd(LBkLegOffset+KneeOffset, RAD(backLeftKnee));

    mseq_mc->setOutputCmd(RBkLegOffset+RotatorOffset, RAD(backRightRotator));
    mseq_mc->setOutputCmd(RBkLegOffset+ElevatorOffset, RAD(backRightAbductor));
    mseq_mc->setOutputCmd(RBkLegOffset+KneeOffset, RAD(backRightKnee));
  }


  virtual void processEvent(const EventBase& event) {
    if ((event.getGeneratorID()) == EventBase::motmanEGID && event.getTypeID() == EventBase::deactivateETID) {
      postCompletionEvent(0);
    } else if (event.getGeneratorID() == EventBase::textmsgEGID) {

      const TextMsgEvent &txtev = dynamic_cast<const TextMsgEvent&>(event);

      int result = atoi(txtev.getText().c_str());
      std::cout << "Desired Distance: " << result << std::endl;

      SharedObject<MediumMotionSequenceMC> mseq_mc;
      erouter->addListener(this,EventBase::motmanEGID); //look for button input

      // Variable Power Kick
      variableKick(mseq_mc,result);

      motman->addPrunableMotion(mseq_mc);
    } 
  }

  virtual void DoStop() {
    erouter->removeListener(this); //generally a good idea, unsubscribe all
    StateNode::DoStop(); // do this last (required)
  }

  static std::string getClassDescription() { return "Variable Kick Node"; }
  virtual std::string getDescription() const { return getClassDescription(); }



private:
  InvertedKick(const InvertedKick&); //!< don't call (copy constructor)
  InvertedKick& operator=(const InvertedKick&); //!< don't call (assignment operator)
};

#endif
